package person;

import java.util.Iterator;
import java.util.List;

import static org.junit.Assert.*;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.Spy;

import static org.mockito.Mockito.*;

import org.mockito.runners.MockitoJUnitRunner;

import organization.Employee;
import address.Address;

//@FixMethodOrder(MethodSorters.NAME_ASCENDING)
// Above annotation is used to make the test cases to execute in the ascending
// of their names. MethodSorters provide 3 types to sort the methods but it is
// never suggested to write test cases to make them to execute in particular order.
// Test cases must be independent of each other. But in case if have such requirement
// then one can use the above annotation but for this Junit4.6 or above is required.

// You need to use below annotation if want to execute the test cases using mocked object
// & it is one of the way to use the mocked objects. 
@RunWith(MockitoJUnitRunner.class)
public class EmployeeTest {

	//	Using below annotation we can mock the object of a class or interface.
	//	Below I have used the interface.
	@Mock
	Address address;

	@Mock
	Person person;

	@Mock
	List<Person> persons;

	@Spy
	Employee employee;

	//	Below will be required when I don't use annotation as used currently.
	//	Employee employees;

	//	Below annotation is used to execute the method before the execution of
	//	every test case. @BeforeClass is also provided to execute the method once when
	//	the class is loaded.
	//	So depending on the requirement use the annotations.

	@Before
	public void setup(){

		//		Below will be required when I don't use annotation as used currently.
		//		employees = new Employee();

		//		What I am doing here? Nothing, I am just using the concept of spy of Mockito.
		//	    Here I am going to spy on the employee object which I testing also.
		//		I will let all the calls go to original methods except some.
		//		Below will be required when I don't use annotation as used currently.
		//		employee = spy(employees);

		//		Now what is this? This is another way to initialize the mock objects or you
		//		can use the annotation used above the class. Use any method as per your comfort.
		//		MockitoAnnotations.initMocks(this);

		//		Below is another way to mock the required object. mock() is overloaded
		//		to provide different features, check it once. Though it will not impact if use
		//		both @Mock & below way to mock the the objects but why use both ways
		//		to do the same thing.
		//		address = Mockito.mock(Address.class);
		//		Not suggested to mock the iterator or List, rather have the original list having mock objects in it.
		//		Below is just shown for information. As such way is fine till
		//		you just want to have 1 or 2 mock objects in your mock list
		//		else if you want to have multiple mock object in your list then
		//		rather have real list having multiple mock objects in it.

		@SuppressWarnings("unchecked")
		Iterator<Person> mockIterator = Mockito.mock(Iterator.class);


		//		Mocking the method calls of the mocked object.		
		Mockito.when(address.getCity()).thenReturn("Pune");
		Mockito.when(address.getState()).thenReturn("MH");
		Mockito.when(address.getCountry()).thenReturn("India");
		Mockito.when(address.getPersons()).thenReturn(persons);
		Mockito.when(persons.iterator()).thenReturn(mockIterator);
		Mockito.when(persons.contains(person)).thenReturn(true);
		Mockito.when(mockIterator.hasNext()).thenReturn(true).thenReturn(false);
		Mockito.when(mockIterator.next()).thenReturn(person);
		//		Also check that I have not used like above Mockito.when() because I have
		//		the import static & now using method directly rather using fully qualified call.
		when(person.getName()).thenReturn("Nitin");
		//		If I uncomment the below statement then my test case testVerifyAddress() will fail
		//		because there I am expecting to get the name as "Nitin" but below statment value
		//		will take the priority over previous statement.
		//		when(person.getName()).thenReturn("Neetant");

		//		To mock a method to throw an exception.
		//		I had to use doThrow() instead of when() because when() can only take
		//		those methods which return some value.
		//		Comment below statement if don't want to throw the exception & test the exception.  
		//		doThrow(new NullPointerException()).when(address).setPersons(persons);

		//		For below method call I could have used doThrow method but used to show that
		//		we can use only those methods in when() which return some value.
		//		Comment below statement if don't want to throw the exception & test the exception. 
		//		when(address.getPersons()).thenThrow(new ArrayIndexOutOfBoundsException());

		employee.getIndividual("1").setAddress(address);

		//		Below are the methods of the class under test, for which calls will be intecepted as below.
		doReturn(true).when(employee).isActive();
		doReturn(person).when(employee).getIndividual("123");
	}

	//	Below is the test case annotated using @Test
	//	@Test further provides 2 properties which can be used if required - expected & timeout

	@Test
	public void SetAddress() {
		//		if the setup() needs to be executed before the execution of every test case then better
		//		to use the annotation @Before. If only once then can execute the below statement.
		//		setup();
		assertNotNull(address.getCountry());
		//		Why I did use verify() below, I am just verifying if any method of Address class
		//		was called in current test case. 
		verify(address).getCountry();
		employee.getIndividual("123").setAddress(address);
	}

	@Test
	public void testGetAddress() {
		//		If the setup method is not executed here before the get call on employee object then
		//		will get the NullPointerException because employee object will not get initialized.
		//		Even if we put the employee object initialization outside the setup() then still we will
		//		get the NullPointerException, but this time it will be for the get() call on Address
		//		as address object will not be there for employee object due to employee object will be
		//		fresh for this test case though it seems that employee object got the address object
		//		in previous test case.

		//				setup();
		if(employee == null || employee.getAddress() == null){
			System.out.println("Employee is null");
			verify(address, never());
			fail("Employee is null");
		}else{
			assertNotNull(employee.getAddress().getCity());
		}
	}

	@Test
	public void testGetPersons(){
		assertNotNull(employee.getPersons(address));
	}

	@Test
	public void testVerifyAddress() {
		//		Check the import statement & usage below. I have imported the static members of the Assert class.
		//		So I don't need to use like Assert.assertTrue(), I have used the static method directly.
		assertTrue(employee.verifyAddress(address, employee.getIndividual("123")));
	}

	@Test
	public void testIsPersonPresent() {
		assertTrue(employee.isActive());
	}
}
